"""Holds common column names, constant through all Butterfree."""

from typing_extensions import Final

TIMESTAMP_COLUMN: Final = "timestamp"
PARTITION_YEAR: Final = "year"
PARTITION_MONTH: Final = "month"
PARTITION_DAY: Final = "day"
